/*
	browse: support some cd, ls, pwd

	part of DerMixD
	(c)2007/8 Thomas Orgis, licensed under GPLv2
*/

// For browsing.
#include <dirent.h>
#include <sys/stat.h>
// For getcwd and chdir.
#include <unistd.h>
#include <fcntl.h>
#include "browse.hxx"

// Reasonable value may be 100 or 1000
#define PATH_STEP 100

char *get_the_cwd()
{
	size_t bs = PATH_STEP;
	char *buf = (char*) malloc(bs);
	while((buf != NULL) && getcwd(buf, bs) == NULL)
	{
		char *buf2 = (char*) realloc(buf, bs+=PATH_STEP);
		if(buf2 == NULL){ free(buf); buf = NULL; }
		else debug("pwd: increased buffer to %zu", bs);

		buf=buf2;
	}
	return buf;
}

static void list_item(const char *i, vector<string*> *ah)
{
	struct stat fst;
	if(!stat(i, &fst))
	{
		ah->push_back(new string);
		if(S_ISDIR(fst.st_mode)) *ah->back() = "D";
		else *ah->back() = "F";

		*ah->back() += " ";
		*ah->back() += i;
	}
	else error("cannot stat %s", i);
}

void browse(socketeer_data *data)
{
	// files/directories to list are in data->act.strings
	// output goes to data->and
	// data->act helps with formatting
	vector<string*>* ah = new vector<string*>;
	if(data->act.strings.empty()) data->act.strings.push_back(".");
	ah->push_back(new string);
	data->act.begin(*ah->back());
	for(vector<string>::iterator i = data->act.strings.begin(); i != data->act.strings.end(); ++i)
	{
		struct stat fst;
		if(!stat(i->c_str(), &fst))
		{
			if(S_ISDIR(fst.st_mode))
			{
				DIR* the_dir = opendir(i->c_str());
				if(the_dir != NULL)
				{
					if(data->cwd == NULL){ ah->push_back(new string("! Error: cwd")); continue; }
					if(chdir(i->c_str())!=0){ ah->push_back(new string("! Error: chdir")); continue; }

					dirent *entry;
					while((entry = readdir(the_dir)) != NULL)
					{
						debug("dirent %s", entry->d_name);
						list_item(entry->d_name, ah);
					}
					if(chdir(data->cwd)!=0) ah->push_back(new string("! Error: chdir(cwd)"));
					closedir(the_dir);
				}
			}
			else list_item(i->c_str(), ah);
		}
		else error("cannot stat %s", i->c_str());

		if(i+1 != data->act.strings.end())
		{
			sdebug("empty line");
			// an empty line
			ah->push_back(new string);
		}
	}
	ah->push_back(new string);
	data->act.end(*ah->back());
	data->wrdat.writes(ah);
}

void pwd(socketeer_data *data)
{
	string *msg = new string;
	if(data->cwd == NULL) data->act.give_error("unable to get directory", *msg);
	else data->act.msg(string("dir: ") + data->cwd, *msg);

	data->wrdat.write(msg);
}

void change_dir(socketeer_data *data)
{
	string *msg = new string;
	if(data->act.strings.empty()) data->act.give_error("I need a directory!", *msg);
	else
	{
		if(chdir(data->act.strings[0].c_str())!=0) data->act.give_error("cannot change dir", *msg);
		else
		{
			free(data->cwd);
			data->cwd = get_the_cwd();
			data->act.success(*msg);
		}
	}

	data->wrdat.write(msg);
}

// Oh, I am so non-portable;-)
static bool is_absolute(string &path)
{
	if(!path.empty())
	{
		return (path[0] == '/');
	}
	return false;
}

void adjust_path(socketeer_data *data, string &path)
{
	// First, if it is an ordinary file, it should exist.
	// If it does not exist, I assume it's some URL; something I cannot path-adjust anyway.
	struct stat fst;
	if(!stat(path.c_str(), &fst))
	if(!is_absolute(path))
	if(data->cwd != NULL)
	path = string(data->cwd) + string("/") + path;

	debug("adjusted path: %s", path.c_str());
}

// Just read... no games with seeking and hoping that that caches.
off_t read_file(const string &name)
{
	off_t length = 0;
	char buffer[4096];
	int fd = open(name.c_str(), O_RDONLY);
	if(fd < 0) return -1;
	do
	{
		errno = 0;
		ssize_t got = read(fd, buffer, 4096);
		if(got < 0)
		{
			if(errno != EINTR)
			{
				close(fd);
				return -1; // else: just continue reading
			}
		}
		else
		{
			length += got;
			if(got < 4096) break;
		}

	} while(1);

	close(fd);
	return length;
}
