/*
	comm_data: simple return value/error back transport with semaphore for notifying initiator of action
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#define ME "comm_data"

comm_data::comm_data()
{
	construct(ME);
	sdebug("comm_data full init");
	waitress = new sem_t;
	mysem = true;
	if(sem_init(waitress, 0 , 0) != 0) 
	{
		serror("Cannot initialize semaphore! I'm really scared and backing out.");
		exit(1); //that's drastic! What better to do? Check for sanity? Exceptions?
	}
	errcode = 0;
	errtext = "";
	retval = 0;
	retval_ulong = 0;
	thread = NULL;
}

comm_data::comm_data(sem_t *s) //an already existing and initialized semaphore
{
	construct(ME);
	sdebug("comm_data init with existing semaphore");
	waitress = s;
	mysem = false;
	errcode = 0;
	errtext = "";
	retval = 0;
	retval_ulong = 0;
	thread = NULL;
}

comm_data::comm_data(pthread_t * t)
{
	construct(ME);
	sdebug("comm_data init with existing thread");
	waitress = new sem_t;
	mysem = true;
	if(sem_init(waitress, 0 , 0) != 0) 
	{
		serror("Cannot initialize semaphore! I'm really scared and backing out.");
		exit(1);
	}
	clear();
	thread = t;
}

comm_data::~comm_data()
{
	destbegin(ME);
	if(mysem)
	{
		debug("destroying semaphore %p", (void*) waitress);
		sem_destroy(waitress);
		delete waitress;
	}
	destend(ME);
}

void comm_data::clear()
{
	errcode = 0;
	errtext = "";
	retval = 0;
	retval_ulong = 0;
	retlines.clear(); // Make sure that there aren't valid pointers left (leaking!).
}
