/*
	communicator: all the server stuff (port watcher and socket communicator)
	
	part of DerMixD
	(c)2004-2008 Thomas Orgis, licensed under GPLv2
	
	depends on netcomm to provide actual socket reading/writing
	This needs a nicer rewrap into some class (merge construct/destruct code in mixer in here).
*/

#include "common.hxx"
#include "version.hxx"

#include <signal.h>
#include <sys/resource.h>

#include "comparse.hxx"
#include "netcomm.hxx"

#include "communicator.hxx"
#include "secsleep.hxx"
#include "browse.hxx"
#include "peers.hxx"

#include <sstream>


extern bool globalflag_oldstyle;

spies bnd;
peers pee;
vector<socketeer_data*> socketeers;
pthread_mutex_t socketeers_lock = PTHREAD_MUTEX_INITIALIZER;
vector<pthread_t*> deadites;
pthread_mutex_t deadites_lock = PTHREAD_MUTEX_INITIALIZER;
sem_t *reapsem_p = NULL; // HACK!
watchers_data wdat;

// =====================================
// Socketeer data class
// =====================================

socketeer_data::socketeer_data(int s, vector<action*> * ac, pthread_mutex_t * aclock, sem_t * sac)
{
	construct("socketeer_data");
	sock = s;
	actions = ac;
	actions_lock = aclock;
	semac = sac;
	cwd = get_the_cwd();
	sb = new socket_buffer(sock);
	cd = new comm_data;

	act.def_world = external_api;
	act.swr = &wrdat;
	post_act.swr = &wrdat;
	notify_act.swr = &wrdat;
	act.comdat = cd;
	post_act.comdat = cd;
	notify_act.comdat = cd;
	scratch[SCRATCH_CHARS] = '\0';

	prepare();
	die_insane_bitch = 0;
};

socketeer_data::~socketeer_data()
{
	destbegin("socketeer_data");
	xdebug("[socketeer_data %i] deleting sb at %p", sock, (void*)sb);
	delete sb; //closes socket, too
	xdebug("[socketeer_data %i] deleting cd at %p", sock, (void*)cd);
	delete cd;
	free(cwd);
	destend("socketeer_data");
};

void socketeer_data::prepare()
{
	ans = "";
	//clear the actions
	act.clear();
	post_act.clear();
	notify_act.clear();
	//clear the communication data
	if(!cd->retlines.empty())
	{
		for(vector<tstring*>::iterator i=cd->retlines.begin(); i!=cd->retlines.end(); ++i)
		delete *i;
	}
	cd->clear();
	//postaction definition and answer sting
	postaction = NULL;
	do_post = false;
	do_notify = false;
};

// =====================================
// Socket watchers class
// =====================================

void* socket_watcher(void* data);
void* socket_reaper(void* data);

bool watchers_data::init(vector<action*> *act, pthread_mutex_t *act_lock, sem_t *act_sem, sem_t *reap_sem)
{
	actions      = act;
	actions_lock = act_lock;
	semac        = act_sem;
	reapsem      = reap_sem;
	good         = false;
	if(pthread_create(&reapthread, NULL, socket_reaper, this) == 0);
	{
		good = true;
		add_thread(reapthread, "socketeer reaper", &td_reaper);
	}
	return good;
}

void watchers_data::exit(void)
{
	// Always take care for all watchers.
	debug("Reaping %zu watchers.", watchers.size());
	for(vector<watcher_data*>::iterator w = watchers.begin(); w != watchers.end(); ++w)
	{
		pthread_cancel((*w)->thread);
		pthread_join((*w)->thread, NULL);
		delete (*w)->sl;
		delete (*w);
	}
	watchers.clear();
	if(good) // For reaper only if setup worked.
	{
		sdebug("Reaping the reaper.");
		pthread_cancel(reapthread);
		pthread_join(reapthread, NULL);
		good = false;
	}
}

bool watchers_data::add_watcher(watcher_data *wd)
{
	if(! wd->sl->online || pthread_create(&wd->thread, NULL, socket_watcher, wd) != 0)
	{
		delete wd->sl;
		delete wd;
		return false;
	}
	wd->add_thread(wd->thread, "socket watcher", &wd->td);
	watchers.push_back(wd);
	return true;
}

bool watchers_data::add_tcp(const porttype port, const int local)
{
	watcher_data *wd = new watcher_data(this);
	wd->sl = new port_listener(port,local);
	// Move that info to mixer or the thread again!
	if(wd->sl->online)
	cerr << "[socket_watcher] online on port " << port << ", locality mode: " << local << endl;

	return add_watcher(wd);
}

bool watchers_data::add_unix(const string name)
{
	watcher_data *wd = new watcher_data(this);
	wd->sl = new unix_listener(name);
	if(wd->sl->online)
	cerr << "[socket_watcher] online on UNIX socket " << name << endl;

	return add_watcher(wd);
}

size_t watchers_data::thread_info(vector<tstring*> &tl)
{
	// This is the reaper.
	size_t count = thread_container::thread_info(tl);
	// The watcher threads do count.
	for(vector<watcher_data*>::iterator i = watchers.begin(); i != watchers.end(); ++i)
	count += (*i)->thread_info(tl);

	pthread_mutex_lock(&socketeers_lock);
	// I could also include the deadites...
	for(vector<socketeer_data*>::iterator i = socketeers.begin(); i != socketeers.end(); ++i)
	count += (*i)->thread_info(tl);

	pthread_mutex_unlock(&socketeers_lock);
	return count;
}

// =====================================
// Communication threads.
// =====================================

void socket_writer_cleaner(void *p)
{
	debug("[socket_writer %i] cleaning", ((socketeer_data*)p)->sock);
}

//the writer for a socket; associated with a socketeer
void* socket_writer(void* p)
{
	pthread_cleanup_push(socket_writer_cleaner, p);
	socketeer_data* data = (socketeer_data*) p;
	data->td_write.thread_workout(0);
	while(true)
	{
		safe_sem_wait(&data->wrdat.semo);
		// OK, a message (or many) arrived... or just the kind order to piss off.
		if(data->wrdat.is_end)
		{
			debug("[socket_writer %i] I'm writing myself off.", data->sock);
			break;
		}
		// Get the new stuff.
		data->wrdat.prepare_flush();
		// Now flush it, handing it over to spies while at that.
		// But if there are no spies active, just flush, without the locking stuff.
		if(bnd.active)
		{
			bnd.reading();
			data->wrdat.flush(data->sb, &bnd);
			bnd.done_reading();
		}
		else data->wrdat.flush(data->sb);
	}
	pthread_cleanup_pop(1);
	pthread_exit(NULL);
}

void socketeer_cleaner(void* p);
void* socketeer(void* p);

//some textual errors for the client
string stringerror(unsigned char errcode)
{
	switch(errcode)
	{
		case DEVERR: return "device error";
		case LOADERR: return "error loading resource";
		case BUSY: return "busy";
		case UNKNOWN_COMMAND: return "What are you talking about?";
		case BAD_CHANNEL: return "invalid channel no.";
		case IN_FOLLOW_ANOTHER: return "have another leader";
		case IN_FOLLOW_INVALID: return "no worthy follower";
		case NONSENSE: return "command doesn't fit here";
		case NO_DEVICE: return "could not create device";
		default:
		{
			ostringstream osr;
			osr << "code " << (int) errcode;
			return osr.str();
		}
	}
}

// Wait for connections on a socket and spawn socketeers to handle them.
void* socket_watcher(void* data)
{
	watcher_data *d = (watcher_data*)data;
	d->td.thread_workout(2);

	while(1)
	{
		//set cancellation point here...
		d->sock = d->sl->accept_connection();
		pthread_testcancel();
		if(d->sock >= 0)
		{
			pthread_mutex_lock(&socketeers_lock);
			socketeers.push_back(new socketeer_data(d->sock, d->mum->actions, d->mum->actions_lock, d->mum->semac));
			// create the writer first, as socketeer depends on it
			pthread_create(&socketeers.back()->wrdat.thread, NULL, socket_writer, socketeers.back());
			pthread_create(&socketeers.back()->thread, NULL, socketeer, socketeers.back());
			socketeers.back()->add_thread(socketeers.back()->thread, "socket reader", &socketeers.back()->td_read);
			socketeers.back()->add_thread(socketeers.back()->wrdat.thread, "socket writer", &socketeers.back()->td_write);
			pthread_mutex_unlock(&socketeers_lock);
		}
	}
	pthread_exit(NULL);
}

// This happens only one at a time!
// I need to be sure that all currently known socket_writers are indeed alive!
void reap_one_socketeer(socketeer_data *sd)
{
	// Going to possibly reap a spy, so be safe here and eliminate it first.
	// I want to stop other from sending messages to a dead spy...
	bnd.eliminate(&sd->wrdat);
	pee.eliminate(&sd->wrdat);

	// 1. cancel and join the socketeer... it is a source of writer work.
	sem_cancel(sd->thread, sd->cd->waitress);
	debug("joining socketeer %i", sd->sock);
	if(pthread_join(sd->thread, NULL) != 0) error("bad join on %li", (long)sd->thread);
	else debug("joined socketeer %i", sd->sock);

	// 2. Tell the writer to quit.
	sd->wrdat.end();
	debug("joining socket_writer %i", sd->sock);

	// 3. Wait for it to end.
	if(pthread_join(sd->wrdat.thread, NULL) !=0) error("bad join on %li", (long)sd->wrdat.thread);
	else debug("joined socket_writer %i", sd->sock);
	// Flushing of messages is not needed as no one should have been sending messages after the reap was triggered.

	// 4. delete the data
	delete sd;
}

void reap_socketeer(pthread_t *deadite)
{
	pthread_mutex_lock(&deadites_lock);
	debug("pushing %li on deadites list", (long)(*deadite));
	deadites.push_back(deadite);
	sem_post(reapsem_p);
	pthread_mutex_unlock(&deadites_lock);
}

void reap_all_socketeers()
{
	sdebug("reaping all socketeers");
	pthread_mutex_lock(&socketeers_lock);
	bnd.eliminate_all();
	pee.eliminate_all();
	pthread_mutex_lock(&deadites_lock);
	deadites.clear();
	pthread_mutex_unlock(&deadites_lock);
	for(vector<socketeer_data*>::iterator s = socketeers.begin(); s != socketeers.end(); ++s)
	{
		debug("reaping socketeer %i", (*s)->sock);
		reap_one_socketeer(*s);
	}
	socketeers.clear();
	pthread_mutex_unlock(&socketeers_lock);
}

void* socket_reaper(void* data)
{
	((watchers_data*)data)->td_reaper.thread_workout(2);
	while(42)
	{
		safe_sem_wait(reapsem_p);
		pthread_mutex_lock(&deadites_lock);
		// There is a list of socketeers to be wiped out properly; without zombies!
		// The key is the socketeer thread id.
		for(vector<pthread_t*>::iterator deadite = deadites.begin(); deadite != deadites.end(); ++deadite)
		{
			// find the entry in socketeers list, store it
			socketeer_data *sd = NULL;
			pthread_mutex_lock(&socketeers_lock);
			for(vector<socketeer_data*>::iterator s = socketeers.begin(); s != socketeers.end(); ++s)
			{
				if((*s)->thread == *(*deadite)) // this looks insane
				{
					debug("[reaper] found deadite %li (socket %i)", (long)((*s)->thread), (*s)->sock);
					sd = *s;
					socketeers.erase(s); // only the offical entry is deleted, not the data
					break;
				}
			}
			pthread_mutex_unlock(&socketeers_lock);
			// now we have the data in private
			if(sd != NULL) reap_one_socketeer(sd);
			else error("[reaper] wanted to delete thread %li, which was not to be found in socketeer list!", (long)(*(*deadite)));
		}
		deadites.clear();
		sdebug("[reaper] done for now");
		pthread_mutex_unlock(&deadites_lock);
	}
	pthread_exit(NULL);
}

void socketeer_cleaner(void* p)
{
	#ifdef DEBUG
	int s = ((socketeer_data*)p)->sock;
	debug("[socketeer %i] cleaning...", s);
	#endif
}

/*
	socket communication handler
	
	here we get commands, parse them, then pass to the mixer (possibly waiting for result from different thread) and normally respond to client
*/
void* socketeer(void* p)
{
	pthread_cleanup_push(socketeer_cleaner, p);

	socketeer_data* data = (socketeer_data*) p; //will this variable actually be deleted?
	data->cd->thread = &data->thread; //me

	debug("[socketeer %i] startup", data->sock);
	data->td_read.thread_workout(0);

	//hello
	data->wrdat.write_copy("[connect] DerMixD v" + version);

	while(1) //run endlessly until close/shutdown or read error
	{debug("[socketeer %i] alive", data->sock);
		//get command string from socket
		
		data->ms = "";
		if((data->erg = data->sb->read_line(data->ms)) < 0)
		{
			error("[socketeer %i] Error on reading: %i, faking close command", data->sock, data->erg);
			data->ms = "close";
		}
		
		if(data->ms != "") //got sth.
		{
			debug("[socketeer %i] got msg: %s", data->sock, data->ms.c_str());
			if(bnd.active)
			{
				bnd.reading();
				bnd.hand_over(data->sb->spyinprefix + data->ms, &data->wrdat);
				bnd.done_reading();
			}

			/*
				parsing and processing of action
				
				result of this all is the return message in data->ans (besides the actual effect)
			*/

			//parse command, including full expansion of script into subactions
			if(data->act.parse(data->ms))
			{
				/*
					there are some possibilities...
					
					let's summarize:
						1. actions with SOCK_LOCAL require no communication with mixer and are handled here
						2. SOCK_POST indicates that there is some postprocessing to do (possibly a postaction)
						3. SOCK_NOTIFY adds an appropriate notify action for mixer
						4. plain normal actions... just give the parsed action to mixer
						
						post and notify can be combined
				*/
				// Path filtering just happens...
				if(data->act.def->flags & SOCK_PATH && !data->act.strings.empty())
				adjust_path(data, data->act.strings.back());

				if(data->act.def->flags & SOCK_POST) //do postprocessing / choose postaction
				{
					switch(data->act.def->id)
					{
						case IN_VOLUME: data->act.floats[0] /= 100; break;
						case IN_LOAD: //play command = load with start 
						case IN_INLOAD:
							data->postaction = &in_start;
						break;
						case OUT_LOAD: //play command = load with start 
							data->postaction = &out_start;
						break;
						case IN_SCRIPT: //only the normal one-time script needs postprocessing (adding 1)
							data->act.ints.push_back(1);
						break;
						case ADD_IN:
						case ADD_OUT:
							//there is the nickname as optional argument, mixer needs something
							if(data->act.strings.size() == 0) data->act.strings.push_back("noname");
						break;
						//QUIT and CLOSE: socketeer thread will be cancelled during safe_sem_wait()
						case CLOSE:
							data->act.msg("Goodbye... hope you'll come around again for some funky mixing!",data->ans);
							data->wrdat.write_copy(data->ans);
						break;
						case QUIT:
							data->act.msg("This is the end, my friend!",data->ans);
							data->wrdat.write_copy(data->ans);
						break;
						default: error("[socketeer %i] dunno what to postprocess on action %s (programming error)!", data->sock, data->act.def->name);
					}
				}
				
				//Action! 
				if(!(data->act.def->flags & SOCK_LOCAL)) //mixer handles it
				{
					/*
						Push the main action to mixer and wait.
						
						In case of QUIT and CLOSE commands, the socketeer will be cancelled when waiting.
						Otherwise, data->cd will contain an error code indicating the success of the action.
						In case of error, there is also (possibly) a textual description
					*/
					//lock, push, post, unlock, wait
					debug("[socketeer %i] pushing action", data->sock);
					pthread_mutex_lock(data->actions_lock);
					data->actions->push_back(&data->act);
					sem_post(data->semac);
					#ifdef DEBUG
					int semval;
					sem_getvalue(data->cd->waitress, &semval);
					#endif
					debug("[socketeer %i] posted (sem=%i)!", data->sock, semval);
					pthread_mutex_unlock(data->actions_lock);
					#ifdef DEBUG
					int waitr =
					#endif
					safe_sem_wait(data->cd->waitress); //wait until mixer or channel thread has processed the action
					#ifdef DEBUG
					sem_getvalue(data->cd->waitress, &semval);
					#endif
					// under gdb(?), cancelling just doesn't always work here
					if(data->die_insane_bitch)
					{
						error("[socketeer %i] I should be dead here! Why didn't I get a proper cancel??", data->sock);
						break;
					}
					debug("[socketeer %i] waiting done (sem=%i, waitr=%i), postprocessing...", data->sock, semval, waitr);
					//prepare notification; always
					if(data->act.def->flags & SOCK_NOTIFY)
					{
						debug("[socketeer %i] notify flag set", data->sock);
						data->do_notify = true;
						data->notify_act.def = ((data->act.def->flags & SOCK_IN) ? &in_notify : ((data->act.def->flags & SOCK_OUT) ? &out_notify : &notify));
						data->notify_act.set_sub(&data->act);
					}
					//prepare postaction; only when first action succeded
					if(data->postaction != NULL && (data->cd->errcode == 0))
					{
						debug("[socketeer %i] post flag set", data->sock);
						data->do_post = true;
						data->post_act = data->act;
						data->post_act.def = data->postaction;
					}

					if(data->do_notify || data->do_post)
					{
						debug("[socketeer %i] pushing notify/post action (do_notify=%d; do_post=%d)", data->sock, data->do_notify, data->do_post);
						pthread_mutex_lock(data->actions_lock);
						if(data->do_notify) data->actions->push_back(&data->notify_act);
						if(data->do_post) data->actions->push_back(&data->post_act);
						sem_post(data->semac);
						pthread_mutex_unlock(data->actions_lock);
						safe_sem_wait(data->cd->waitress);
						debug("[socketeer %i] first postwait done", data->sock);
						if(data->do_notify && data->do_post) safe_sem_wait(data->cd->waitress);
						debug("[socketeer %i] second (if any) postwait done", data->sock);
						data->notify_act.set_sub(NULL); //don't try to delete main action on clear!
					}

					/*
						processing of return value either of:
						
						- single action
						- notify
						- postaction
						
						In case of notify _and_ postaction, the return value of notify is being overwritten.
						Boldly requesting that there is some sense in bundling the postaction to the action needing a notify, it is assumed that the postaction will fail also when the notify failed.
					*/
					if(!(data->act.def->flags & SOCK_EXT)) //if socketeer shall give answer
					{
						if(data->cd->errcode == 0) //successful?
						{
							if(!data->cd->retlines.empty())
							{ // Someone prepared answers to give! Pack +begin and -end around and out with it!
								vector<string*> *st = new vector<string*>;
								if(data->act.def->id == THREADSTAT)
								{ // Special: mixer started it... I finish.
									data->cd->retval_ulong += wdat.thread_info(data->cd->retlines);
									data->cd->retlines.push_back(new tstring("thread count: "));
									data->cd->retlines.back()->sprintf("%lu", data->cd->retval_ulong);
								}
								else
								{
									st->push_back(new string);
									data->act.begin(*st->back());
								}
								tstring_v_string(data->cd->retlines, *st);
								data->cd->retlines.clear();
								st->push_back(new string);
								data->act.end(*st->back());
								data->wrdat.writes(st);
								// SOCK_ANS shouldn't be set for this... else a bogus data->ans will be created.
							}
							else data->act.success(data->ans); //create basic success msg

							if(data->act.def->flags & SOCK_ANS) //append some specific info?
							{
								data->ans +=  ": ";
								switch(data->act.def->id)
								{
									case IN_LENGTH:
										snprintf(data->scratch, SCRATCH_CHARS, "%0.3f", data->cd->retval);
										data->ans += data->scratch;
										data->ans += "s; ";
										snprintf(data->scratch, SCRATCH_CHARS, "%lu", data->cd->retval_ulong);
										data->ans += data->scratch;
										data->ans += " samples; ";
										snprintf(data->scratch, SCRATCH_CHARS, "%lu", data->cd->retval_ulong2);
										data->ans += data->scratch;
										data->ans += "kHz";
									break;
									default:
										snprintf(data->scratch, SCRATCH_CHARS, "%0.3f", data->cd->retval);
										data->ans += data->scratch;
								}
							}
						}
						else data->act.give_error(stringerror(data->cd->errcode) + " <- " + data->cd->errtext,data->ans);
					}
				}
				else //local handling
				{
					switch(data->act.def->id)
					{
						case ID: data->act.msg("DerMixD v" + version + "." + subversion, data->ans); break;
						case FADEOUT: data->act.give_error("um, ah... I don't do this... sorry. Use scripting instead", data->ans); break;
						case HELP:
							if(! data->act.strings.empty())
							{
								data->act.msg("",data->ans);
								action_help(find_action_def(data->act.strings[0], external_api),data->ans);
							}
							else
							{
								data->act.msg("usage: help command; available commands: ",data->ans);
								list_actions(external_api, data->ans);
							}
						break;
						case SHOWAPI:
						{
							data->ans = "";
							vector<string*>* ah = new vector<string*>;
							ah->push_back(new string);
							data->act.begin(*ah->back());
							ah->push_back(new string("Full API listing:"));
							ah->push_back(new string(""));
							action_allhelp(external_api, *ah);
							ah->push_back(new string);
							data->act.end(*ah->back());
							data->wrdat.writes(ah);
						}
						break;
						case LS:  browse(data); break;
						case PWD: pwd(data);    break;
						case CD:  change_dir(data);    break;
						case SPY:   bnd.send(&data->wrdat);       data->act.success(data->ans); break;
						case UNSPY: bnd.eliminate(&data->wrdat);  data->act.success(data->ans); break;
						case SHOWID: data->act.idmask = data->act.ints[0] == 0 ? 0 : SOCK_IN|SOCK_OUT; data->act.success(data->ans); break;
						case PEER:
							/* [peer from to] */
							data->act.fullname += " ";
							data->act.fullname += data->act.strings[0];
							data->act.fullname += " ";
							data->act.fullname += data->act.strings[1];
							if(pee.message(data->act))
							data->act.success(data->ans);
							else data->act.give_error("No such peer registered.", data->ans);
						break;
						case ADDPEER:
							if(pee.add(&data->wrdat, data->act.strings[0], data->act.strings[1]))
							data->act.success(data->ans);
							else data->act.give_error("Peer name already taken?", data->ans);
						break;
						case REMPEER:
							if(pee.remove(data->act.strings[0]))
							data->act.success(data->ans);
							else data->act.give_error("No such peer registered.", data->ans);
						break;
						case SHOWPEERS:
							data->ans = "";
							pee.list(data->act, data->wrdat);
						break;
						case IN_PREREAD:
						{
							off_t bytes = read_file(data->act.strings.back());
							if(bytes >= 0)
							{
								ostringstream os;
								os << bytes;
								data->act.success(data->ans);
								data->ans += ": ";
								data->ans += os.str();
								data->ans += " bytes read";
							}
							else data->act.give_error("some problem", data->ans);
						}
						break;
						default: data->act.give_error("socketeer does not know what to do with that local action",data->ans);
					}
				}
			}
			else data->act.success(data->ans); //action failed to parse... success=outcome (old style;-)

			if(data->ans != "") data->wrdat.write_copy(data->ans);
			data->prepare(); //be ready for next round
		} //end of command message processing
		else debug("[socketeer %i] got empty string", data->sock);
		pthread_testcancel();
		
	} //end of endless loop
	pthread_cleanup_pop(1);
	pthread_exit(NULL);
}
