#part of ThOrMA (but not necessarily depending upon)
#DerMixD::Base: basic interface to dermixd

#Copyright (c) 2005/6 Thomas Orgis (thomas@orgis.org)
#This is Open Source, Distribution and Modification under the terms of the Artistic License
#see the file LICENSE in the ThOrMA package root or at a place some steps upwards directory-wise if you got this file in a separate package

#This is the bad basic module... it uses old-style Socket with old-style filehandle... not sure, but I think this could be very bad when trying to make more than one connection (creating more than one object) in one script. Nevertheless, it compiles way faster that with full oo'ed Socket access.

package DerMixD::Base;

use strict;
use Socket;

our %long =
(
	'g', 'getstat',
	'f', 'fullstat',
	's', 'shutdown'
);

our %default =
(
	'remote', 'localhost',
	'port', '8888',
	'debug', 0,
	'chan', 0,
	'echo', 0
);

#hack for my old OSF1 Perl:
#don't like to give up quotes!
my $q = '\Q';
my $e = '\E';
if(`uname` =~ /^OSF1/){ $q = ''; $e = ''; }
#damn! that never works! ... There is something wrong with my understanding of regex building here
$q = ''; $e = '';

my $me = '[DerMixD::Base]';
#talk with some dermixd-1.0+ 
my $major = 1;
my $miniminor = 0;

sub new
{
	my $class = shift;
	my $self = shift; #remote, <host>, port, <port>, ...
	for(keys %default){ $self->{$_} = $default{$_} unless defined $self->{$_} }
	$self->{online} = 0;
	bless $self, $class;
	$self->Open();
	return $self;
}

#gets full status into internal structured hash
sub FullStatus
{
	my $self = shift;
	@{$self->{in}} = ();
	@{$self->{out}} = ();
	for(@{$self->Command(['fullstat'])})
	{
		#in 0:Lena -> (0) stopped at 0s/150.152s (buf: 0) vol: 1 speed: 1 eq: 1|1|1, dev: mp3 file: /home/thomas-data/thorma/var/music/phantoms_of_futur
		#in 1:Elena -> (0) idle at 0s/0s (buf: 0) vol: 1 speed: 1 eq: 1|1|1
		#out 0:Helena <- (0,1) playing dev: oss file: /dev/dsp
			if(/^in\s+(\d+):(\S+)\s*->\s*(\([\d,]*\))\s+(\S+)\s+at\s+([\d\.]+)s\/([\d\.]+)s\s+\([^\(]+\)\s+vol:\s*([\d\.]+)\s+speed:\s*([\d\.]+)\s+eq:\s*([\d\.]+)\|([\d\.]+)\|([\d\.]+)\s*(\,\s*following\s+(\d+:[^,]*)|)\s*(\,\s*(\d+)\s*actions|)\s*(\,?\s*dev:\s*(\S+)\s+file:\s*(.*)|)$/)
		{
			next if $1 < 0;
			$self->{in}->[$1] = 
			{
				'num', $1,
				'nickname', $2,
				'status', $4,
				'position', $5,
				'length', $6,
				'volume', $7,
				'speed', $8,
				'bass', $9,
				'mid', $10,
				'treble', $11,
				'actions', defined $15 ? 1*$15 : 0
			};
			#targets from $3 (better reference list?)
			@{$self->{in}->[$1]->{targets}} = split(',', $3);
			
			if($16)
			{
				$self->{in}->[$1]->{device} = $17;
				$self->{in}->[$1]->{file} = $18;
			}
		}
		elsif(/^out\s+(\d+):(\S+)\s*<-\s*(\([\d,]*\))\s+(\S+)(\s+dev:\s*(\S+)\s+file:\s*(.*)|)$/)
		{
			next if $1 < 0;
			$self->{out}->[$1] = 
			{
				'num', $1,
				'nickname', $2,
				'status', $4
			};
			#sources from $3 (better reference list?)
			@{$self->{out}->[$1]->{sources}} = split(',', $3);
			if($5)
			{
				$self->{out}->[$1]->{device} = $6;
				$self->{out}->[$1]->{file} = $7;
			}
		}
		else{ print STDERR "$me not understanding this fullstat line: $_\n"; }
	}
}

#issue a command and wait for / parse response
sub Command
{
	my $self = shift;
	my $com = shift;
	my @l; #answer lines
	$com->[0] = $long{$com->[0]} if defined $long{$com->[0]};
	unless($com->[0] eq 'getstat')
	{
		my $in = '';
		print STDERR "$me >@{$com}\n" if $self->{debug};
		print SOCK "@{$com}\n";
		while(defined (my $line = <SOCK>))
		{
			print $line if $self->{echo};
			$line =~ s/\r*\n*$//; #really chomp!
			unless($in)
			{
				#ignoring lines not belonging to current request
				if($line =~ /^(\[\Q$com->[0]\E(|\/.+)\])\s*(.*)$/)
				{
					if(length($3) >= 1 and substr($3,0,1) eq '+'){ $in = $1; }
					else{ push(@l, $3); last; }
				}
			}
			else
			{
				if($line =~ /^\[\Q$com->[0]\E(|\/.+)\]\s*-/){ last; }
				else{ push(@l, $line); }
			}
			print STDERR "$me still waiting for $com->[0]\n" if $self->{debug};
		}
		if($com->[0] eq 'shutdown' or $com->[0] eq 'close'){ $self->{online} = 0; }
		
		#unless($l){die "Somehow the command $com did not get a response... I'm scared. ($!)\n";}
		print STDERR "$me <".join("\n$me <",@l)."\n" if $self->{debug};
	}
	else{ push(@l, "$me refrains from using this deprecated command."); }
	return \@l;
}

#wait for a specific string as produced by a timed say <string> command
sub WaitForSay
{
	my $self = shift;
	my $word = shift;
	if($self->{online})
	{
		#1.1 introduced [saying] before the string...
		#Tru64 perl has problem with \Q \E apparently
		my $regex = ($self->{interface} eq '1.0') ? '^'.$q.$word.$e.'\r*\n*$' : '^\[saying\]\s+'.$q.$word.$e.'\r*\n*$';
		while(defined (my $line = <SOCK>))
		{
			print STDERR "waitforsay: got $line";
			return 1 if $line =~ $regex;
		}
	}
	return 0;
}

#Make connection to the daemon
#attention: I think SOCK is treated as class handle here... so no several connections at once, thus OpenSocket closes an existing connection before opening new...(?)
sub Open
{
	my $self = shift;
	print STDERR "$me want to connect to host $self->{remote}:$self->{port}\n" if $self->{debug};
	
	$self->Close();
	
	#get the network on
	my $iaddr = inet_aton($self->{remote}) or return 0;
	socket(SOCK, PF_INET, SOCK_STREAM, getprotobyname('tcp')) or return 0;
	connect(SOCK, sockaddr_in($self->{port}, $iaddr)) or return 0;
	my $oldhandle = select(SOCK); $| = 1; select($oldhandle); # flushing before things starts stinking
	
	#read and interpret greeting
	my $greeting = <SOCK>;
	if($greeting =~ /^\[connect\]\s*DerMixD\s*v(\d+)\.(\d+)/)
	{
		$self->{interface} = "$1.$2";
		$self->{major} = $1;
		$self->{minor} = $2;
		if($self->{major} == $major)
		{
			if($self->{minor} >= $miniminor)
			{
				$self->{online} = 1;
				print STDERR "$me connected to host $self->{remote}\n" if $self->{debug};
				$self->FullStatus();
			}
			else{	print STDERR "$me need at least DerMixD v$major.$miniminor!\n"; }
		}
		else{	print STDERR "$me need DerMixD v$major.x interface (not $self->{interface})!\n"; }
	}
	else
	{
		print STDERR "$me want to know who's there\n" if $self->{debug};
	}
	return $self->{online};
}

#end communication with the current dermixd
sub Close
{
	my $self = shift;
	if($self->{online})
	{
		print SOCK "close\n";
		#to extreme?
		close (SOCK) or die "$me close $!";
		$self->{online} = 0;
	}
}

sub DESTROY
{
	my $self = shift;
	$self->Close();
}

1;
