/*
	in_raw_s16: RAW 16bit (no thoughts about byte order) input device

	part of DerMixD
	(c)2004-7 Thomas Orgis, licensed under GPLv2

	When I support different sample formats than signed 16bit, this input device will be extended...
*/

#include "common.hxx"
#include <fcntl.h>
#include <sys/types.h>
#include <signal.h>
#include "debug.hxx"
#include "audio_buffer.hxx"
#include "audio_functions.hxx"
#include "input.hxx"
#include "in_raw_s16.hxx"
#include "comparse.hxx"

const string raw_s16_input::t = "raw_s16";

raw_s16_input::raw_s16_input(input_data* indat)
{
	construct("raw_s16_input");
	data = indat;
	type = &t;
	on_line = true;
	im_ready = false;
}

raw_s16_input::~raw_s16_input()
{
	destbegin("raw_s16_input");
	if(im_ready) file.close();
	destend("raw_s16_input");
}

int raw_s16_input::load(comm_data* cd, const string& filename)
{
	if(im_ready)
	{
		sdebug("closing old file...");
		file.close();
		im_ready = false;
	}
	debug("trying to open %s", filename.c_str());
	file.open(filename.c_str(), ios::in|ios::binary|ios::ate);
	if(file.is_open())
	{
		sdebug("success");
		//its easy here to get the file parameters 
		data->length = (unsigned long) file.tellg() / (data->format.channels*sizeof(audio_type));
		file.seekg (0, ios::beg);
		data->position = 0;
		data->filename = filename;
		im_ready = true;
		data->at_end = false;
		data->near_end = false;
	}
	sem_post(cd->waitress);
	return ( im_ready ? 0 : -1 );
}

int raw_s16_input::seek(comm_data* cd, double position)
{
	return sampleseek(cd, samples(&(data->format), position));
}

 //try to seek to data->position, set it to really achieved value, respond to socket
int raw_s16_input::sampleseek(comm_data* cd, unsigned long position)
{
	int ret = 0;
	if(im_ready)
	{
		debug("seeking to %lu", position);
		// this fstream sux - seekg doesn't clear eof! In fact, it does nothing on eof...
		if(file.eof()) file.clear();
		file.seekg( bytes(&(data->format), position));
		if(file.good())
		{
			sdebug("file still good");
			data->position = position;
			data->at_end = false;
		}
		else
		{
			if(file.eof()){ data->at_end = true; data->position = data->length; }
			else{ im_ready = false; file.close(); ret = -2; }
		}
	}
	else ret = -1;
	if(cd != NULL){ cd->retval = (double) data->position/44100; sem_post(cd->waitress); }
	return ret;
}


int raw_s16_input::play(unsigned int n_samples) //fill buffer with mixer data format (after pitch / resampling)
{
	sdebug("play");
	int ret = -1;
	if(im_ready)
	{
		if(n_samples > buffer.size)
		{
			buffer.reallocate(n_samples);
		}
		if(buffer.size >= n_samples)
		{
			if(!data->at_end)
			{
				debug("play: really reading now at %zu", (size_t) file.tellg());
				unsigned long int start = (unsigned long int) file.tellg();
				file.read((char*) buffer.ptr, (streamsize) bytes(&(data->format), n_samples));
				if(file.eof()) 
				{
					data->at_end = true;
					sdebug("play: got eof");
					buffer.fill = (unsigned int) (data->length - (start / (data->format.channels*sizeof(audio_type))));
				}
				else{ buffer.fill = (unsigned int) (( (unsigned long int) file.tellg() - start )  / (sizeof(audio_type) * data->format.channels)); }
				data->position += buffer.fill;
				ret = 0;
			}
		}
		else ret = -1;
	}
	sem_post(data->insem);
	debug("play: done, buffer fill: %u", buffer.fill);
	return ret;
}

//need real eq...
int raw_s16_input::set_eq(comm_data* cd)
{
	if(cd != NULL) sem_post(cd->waitress);
	return 0;
}

