/*
	in_sndfile: sndfile input device
	
	part of DerMixD
	(c)2007 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "input.hxx"
#include "in_sndfile.hxx"

#define ME "sndfile_input"
const string sndfile_input::t = "sndfile";

snd_file::snd_file()
{
	construct("snd_file");
	file = NULL;
}

snd_file::~snd_file()
{
	destbegin("snd_file");
	do_close();
	destend("snd_file");
}

bool snd_file::do_read(audio_buffer *b, unsigned int wanted_samples)
{
	if(file != NULL)
	{
		b->fill = sf_readf_short(file, (short*)b->ptr, wanted_samples);
		return true;
	}
	else
	{
		b->fill = 0;
		return false;
	}
}


bool snd_file::do_open(input_data *d)
{
	do_close();
	info.format = 0;
	file = sf_open(d->filename.c_str(), SFM_READ, &(info));
	if(file != NULL)
	{
		d->length = info.frames;
		d->format.channels = info.channels;
		d->format.rate = info.samplerate;
		// be prepared for float files read as int (I should work with float in that case...)
		sf_command(file, SFC_SET_CLIPPING, NULL, SF_TRUE);
		sf_command(file, SFC_SET_SCALE_FLOAT_INT_READ, NULL, SF_TRUE);
	}
	return file != NULL;
}

void snd_file::do_close()
{
	if(file != NULL) sf_close(file);
	file = NULL;
}

bool snd_file::do_seek(input_data *d)
{
	if(file != NULL)
	{
		/* length info is accurate... we work with file headers... not appending to files while reading... */
		if(d->position >= d->length) d->position = d->length;
		sf_count_t realpos = sf_seek(file, d->position, SEEK_SET);
		if(realpos >= 0)
		{
			d->at_end = false;
			d->position = realpos;
			return true;
		}
	}

	return false;
}

sndfile_input::sndfile_input(input_data* ind)
{
	construct(ME);
	type = &t;
	build(ind, &sf);
}

sndfile_input::~sndfile_input()
{
	destbegin(ME);
	destroy();
	destend(ME);
}
