/*
	in_vorbisfile: vorbisfile input device (OGG Vorbis files)

	initially written by Sven Oliver Moll

	part of DerMixD
	(c)2007 Sven Oliver Moll and Thomas Orgis, licensed under GPLv2
*/

#include "in_vorbisfile.hxx"
#include "debug.hxx"

#define ME "vorbisfile_input"
#define MEF "vorbis_file"

// Hm, actually we talk about vorbus, ogg is generic... but it is the common file name ending.
// Dammit, I should code real file type detection... using libmagic... test access.
const string vorbisfile_input::t = "ogg";

// The actually meat is the vorbis_file class below, vorbisfile_input is the glue to the input framework.

vorbisfile_input::vorbisfile_input(input_data* ind)
{
	construct(ME);
	type = &t;
	build(ind, &vf);
}

vorbisfile_input::~vorbisfile_input()
{
	destbegin(ME);
	destroy();
	destend(ME);
}

// Down to the real work.

vorbis_file::vorbis_file()
{
	construct(MEF);
	memset(&handle, 0, sizeof(handle));
	opened = false;
}

vorbis_file::~vorbis_file()
{
	destbegin(MEF);
	do_close();
	destend(MEF);
}

void vorbis_file::do_close()
{
	// file gets closed by vorbisfile... (?)
	if(opened) ov_clear(&handle);
	opened = false;
}

bool vorbis_file::do_open(input_data *data)
{
	do_close();
	debug("going to open %s", data->filename.c_str());
	file = fopen(data->filename.c_str(), "rb");
	if(file)
	{
		sdebug("success");
		ov_open(file, &handle, 0, 0);
		vorbis_info *vi = ov_info(&handle, 0);
		//its easy here to get the file parameters
		data->format.rate = vi->rate;
		data->format.channels = vi->channels;
		data->length = ov_pcm_total(&handle,-1);
		opened = true;
		return true;
	}
	return false;
}

bool vorbis_file::do_read(audio_buffer *b, unsigned int n_samples)
{
	int ret = true;
	long decoded = -1;
	long size_left = (long)b->channels*sizeof(audio_type)*n_samples;
	int current_section = 0; // Initialize?
	char *cbuffer = (char*)(&b->ptr[0]);

	if(!opened) return false; // Unnecessary paranoia?
	while(size_left > 0)
	{
		decoded = ov_read(&handle, cbuffer, size_left, 0, 2, 1/*sgned*/, &current_section);
		if(decoded <= 0)
		{
			sdebug("play: got eof");
			break;
		}
		if(current_section != 0) error("current_section changed to %i ... this may include a format change, which I do not support.", current_section);

		cbuffer += decoded;
		size_left -= decoded;
	}
	if(decoded < 0) ret = false;

	sdebug("filled");
	b->fill = n_samples - (size_left/(b->channels*sizeof(audio_type)));
	return ret;
}

bool vorbis_file::do_seek(input_data *data)
{
	int ret = true;

	if(opened)
	{
		if(data->position >= data->length) data->position = data->length;
		int seekret = ov_pcm_seek(&handle, data->position);
		switch(seekret)
		{
			case 0:
			break;
			case OV_ENOSEEK:  // Bitstream is not seekable.
			case OV_EINVAL:   // Invalid argument value; possibly called with an Ogghandle structure that isn't open.
			case OV_EREAD:    // A read from media returned an error.
			case OV_EFAULT:   // Internal logic fault; indicates a bug or heap/stack corruption.
			case OV_EBADLINK: // Invalid stream section supplied to libvorbisfile, or the requested link is corrupt.
			default: ret = false;
		}
	}
	else ret = false;

	return ret;
}
