#ifndef DMD_H_INPUT
#define DMD_H_INPUT

/*
	input: general input device and data definition 
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
	
	This includes the clumsy function that creates an input device based on a string value (its generic name...).
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "thread_container.hxx"
#include "action_queue.hxx"

/* types */

class input_data;

class input_file
{
public:
	string errtext;
	unsigned char errcode;

	input_file(){ errcode = 0; };
	virtual ~input_file();
	void explain_error(comm_data *cd, unsigned char aftercode, const string aftertext);
	virtual bool do_read(audio_buffer *buffer, unsigned int wanted_samples) = 0;
	virtual bool do_open(input_data *data) = 0;
	virtual void do_close() = 0;
	// Better interface is with explicit position. The method overload here is transitional code.
	virtual bool do_seek(input_data *data);
	virtual bool do_seek(input_data *data, unsigned long pos);
	virtual bool do_length(input_data *data) = 0;
	// Silent failure per default...
	virtual bool do_eq(const vector<double> &val){ return true; };
};

class inwork_data: public thread_data
{
	public:
	action_queue worklist;
	input_data *data;
	audio_buffer *buffer;
	input_file *inf;
	/* For preseek. */
	bool new_pos;
	unsigned long the_new_pos;
};


class input_data
{
	public:
		int id;
		volatile double bass,mid,treble; //as linear (?) factors, look at mpg123's eq
		string filename; //not volatile but hopefully too large for a cpu register
		volatile unsigned long position; //internal samples
		volatile unsigned long length; 	//internal samples
		volatile audio_format format;
		volatile bool at_end; //only the play()-thread is allowed to set this! ... pos=length is not enough as this info (length) may not be 
		volatile bool near_end; //backwatch sets this when mpg123 is finished
		//reliable... unset this on seeking to beginning; looking here on play attempt (in main); at end is also triggered when no file there
		sem_t* insem;
		vector<unsigned long int> specials; //like MP3SamplesPerFrame
		input_data(){ reset(); };
		void reset()
		{
			position = length = 0;
			at_end = near_end = false;
			specials.clear();
			filename = "";
		};
};

class input_device: public thread_container
{
	protected:
		input_data* data;
		inwork_data wd;
		volatile bool on_line;
		pthread_t wt;

	public:
		const string *type;
		string problem;
		audio_buffer buffer;

		input_device();
		virtual ~input_device();
		void build(input_data *ind, input_file *inf);
		void destroy();
		virtual int load(comm_data* cd, const string& filename); //load the file stated in chdata, respond to socket
		virtual	int seek(comm_data* cd, double position); //try to seek to data->position, set it to really achieved value, respond to socket
		virtual int sampleseek(comm_data* cd, unsigned long position);
		virtual void scan_length(comm_data *cd); // scan till end found, seek back, return info
		virtual void scan(comm_data *cd, const string &list); // scan through file for various info, seek back
		void no_scan_length(comm_data *cd);      // scan_length when not supported
		void done_scan_length(comm_data *cd);    // scan_length when no extra action needed
		virtual int set_eq(comm_data* cd);
		virtual bool online();
		virtual bool ready(); //for a start
		virtual	int play(unsigned int n_samples); //fill buffer with mixer data format (after pitch / resampling)
};

/* functions */

input_device* new_input_device(string type, input_data& data);

#endif
