#ifndef DMD_H_MIXER_DATA
#define DMD_H_MIXER_DATA

/*
	mixer_data: data types and defines for mixer 
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
	
	integral part are the channel (both in and out with shared code) classes
*/

//status.
#define PLAYING 1
#define PAUSED 2
#define DEAD 0 
#define NODEV 8
#define IDLE 3 //device there and no file loaded
#define STOPPED 4 //there is a file loaded
#define SEEKING_PAUSE 5
#define SEEKING_PLAY 6
#define LOADING 7
#define NOTHING 255

/*
	an output channel knows these states:
	PLAYING: device online and there is data to play
	IDLE: device online nut nothing to play
	NODEV: device not online
	DEAD: really fucked up, should be erased from memory...
 */

#include "thread_container.hxx"

typedef struct timeaction
{
	double time;
	long  count;
	action *act;
} timeaction;

class channel: public thread_container
{
	public:

	bool working;
	bool offline;
	signed char status;
	string name;
	vector<timeaction> script;
	vector<timeaction> postscript;

	virtual void now_and_then(unsigned int chunksize, double &now, double &then) = 0;
	virtual ~channel(){};
	
	bool getactions(unsigned int chunksize, vector<action*> &nowscript);
	vector<timeaction>::size_type addaction(const double time, const long count, action*);
	void delscript();
	void script_nirvana(socket_writer_data* dead_writer);
	bool getpostactions(vector<action*> &nowscript);
	vector<timeaction>::size_type addpostaction(const double time, const long count, action*);
};

class inchannel;

class outchannel: public channel
{
	public:
	output_device* device;
	output_data data;

	vector<inchannel*> sources;

	outchannel(int mid, const string& namen);
	~outchannel();

	unsigned char load(const string type, const string resource, comm_data *comdat);
	bool device_sane();
	void nodev();
	
	//script helper
	void now_and_then(unsigned int chunksize, double &now, double &then);
};

class inchannel: public channel
{
	public:
	input_device* device;
	input_data data;
	mixer_buffer buffer;
	int startval[2];
	bool buffering;
	inchannel* follower;
	inchannel* leader;
	vector<socket_writer_data*> watchers;
	unsigned int hadseconds;
	signed char hadstatus;
	unsigned int ordered_samples;

	double speed;
	double volume;

	vector<outchannel*> targets; //pointers to output objects... just make sure that on deletion of those the pointers here are deleted first (?)

	inchannel(int mid, const string& namen, sem_t* inns);

	~inchannel();

	unsigned char load(string &resource, comm_data *comdat, const string type = "auto");
	bool device_sane();
	void nodev();	
	unsigned long position_in_seconds(){ return (unsigned long) seconds(&data.format, data.position); };
	bool resize_buffer(unsigned int s);
	void reset_buffer();

	//script helper
	void now_and_then(unsigned int chunksize, double &now, double &then);
	void get_script_commands(vector<string*> &list);
};

#endif
