/*
	out_dummy: dummy output device
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "output.hxx"
#include "out_dummy.hxx"
//needed to fake work
#include "secsleep.hxx"

const string dummy_output::t = "dummy";

/*
	Here is where the music plays!
	
	Be sure to keep the workflow:
		endlessly
		wait for "in" semaphore
		do work (nothing wrong with blocking I/O)
		signal on "out" semaphore
*/

void* dummy_writer(void* p)
{
	writerdata* data = (writerdata*) p;
	dummy_writerextra* extra = (dummy_writerextra*) data->extra;
	#ifdef DEBUG_OUTPUT
	cerr << "[dummy_output] writer thread started" << endl; 
	#endif
	while(1)
	{
		safe_sem_wait(&data->insem); //"Gimme some work!"
		data->error = 0; //nothing went wrong yet
		
		#ifdef DEBUG_OUTPUT
		cerr << "[dummy_output] writer thread writing" << endl; 
		#endif
		//do the actual playback action... normally some kind of writing the stuff in data->buffer to some device/file
		secsleep((double) extra->time_factor/extra->format->rate*data->buffer.size);
		
		//set data->error to sth != 0 to indicate some failure
		#ifdef DEBUG_OUTPUT
		cerr << "[dummy_output] writer thread wrote, posting on " << (void*) &data->outsem<<  endl; 
		int semval;
		sem_getvalue(&data->outsem, &semval);
		cerr << "[dummy_output] writer thread semaphore " << (void*) &data->outsem << " has value " << semval <<  endl; 
		#endif
		
		sem_post(&data->outsem); //"Hey, I'm done!"
		
		#ifdef DEBUG_OUTPUT
		sem_getvalue(&data->outsem, &semval);
		cerr << "[dummy_output] writer thread semaphore " << (void*) &data->outsem << " now has value " << semval <<  endl; 
		#endif
		
	}
	pthread_exit(NULL);
}

dummy_output::dummy_output(output_data* odi, unsigned int bufsize)
{
	construct("dummy_output");
	/*
	  build() makes standard init, executes special_init() and starts the writer thread.
		Normally you call only build() here and have special init stuff in special_init().
	*/
	build(odi, bufsize, dummy_writer);
}

/*
	Wonders of C++... there are limits to virtuality
	
	You _must_ call destroy() here and please put all custom destruction stuff into special_exit() unless you _really_ know what you are doing!
*/
dummy_output::~dummy_output()
{
	destbegin("dummy_output");
	destroy();
	destend("dummy_output");
}

bool dummy_output::special_init()
{
	//initialize any special stuff here...
	//...like variables for the writer
	wextra.format = &data->format;
	wextra.time_factor = 0.9; 
	wdat.extra = (void*) &wextra; //add pointer to custom data structure to the writer's data structure
	return true; //return false if sth. failed here
}

void dummy_output::special_exit()
{
	#ifdef DEBUG_MEM
	cerr << "[dummy_output" << data->id << "] leaving..." << endl;
	#endif
	//you would like to deallocate any self-allocated memory... or whatever
	//there is quite a chance for this method staying empty because special_deactivate() is called by parent destructor anyway
}

bool dummy_output::special_activate()
{
	#ifdef DEBUG_OUTPUT
	cerr << "[dummy_output" << data->id << "] activate" << endl; 
	#endif
	//open device node/file, connect to server, ...return success
	return true;
}

void dummy_output::special_deactivate()
{
	#ifdef DEBUG_OUTPUT
	cerr << "[dummy_output" << data->id << "] deactivate" << endl; 
	#endif
	//close device/file/connection
}

void dummy_output::special_pause()
{
	#ifdef DEBUG_OUTPUT
	cerr << "[dummy_output" << data->id << "] pause" << endl; 
	#endif
	//prepare for future silence... workaround for OSS; future of this method being unsure
}
