/*
	out_text: text output device
	
	part of DerMixD
	(c)2005 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "output.hxx"
#include "out_text.hxx"

#include <fcntl.h>

const string text_output::t = "text";

void* text_writer(void* p)
{
	writerdata* data = (writerdata*) p;
	text_writerextra* extra = (text_writerextra*) data->extra;
	#ifdef DEBUG_OUTPUT
	cout << "[text_output] writer thread started" << endl; 
	#endif
	data->thread_workout(0);
	while(1)
	{
		safe_sem_wait(&data->insem);
		data->error = 0;
		
		#ifdef DEBUG_OUTPUT2
		cout << "[text_output] writing\n"; 
		#endif
		data->buffer.print_text(extra->file);
	 
		sem_post(&data->outsem);
	}
	pthread_exit(NULL);
}

text_output::text_output(output_data* odi, unsigned int bufsize)
{
	construct("text_output");
	build(odi, bufsize, text_writer);
}

text_output::~text_output()
{
	destbegin("text_output");
	destroy();
	destend("text_output");
}

bool text_output::special_init()
{
	wdat.extra = (void*) &wextra;
	return true;
}

void text_output::special_exit()
{
	#ifdef DEBUG_MEM
	cout << "[text_output << " << data->id << "] leaving..." << endl;
	#endif
}

bool text_output::special_activate()
{
	bool succ = false;
	wextra.file.clear();
	wextra.file.open(data->filename.c_str());
	if(wextra.file.is_open())
	{	
		#ifdef DEBUG_OUTPUT
		cout << "[text_output " << data->id << "] printing header" << endl; 
		#endif
		buffer.print_text_header(wextra.file);
		succ = true;
	}
	return succ;
}

void text_output::special_deactivate()
{
	#ifdef DEBUG_OUTPUT
	cout << "[text_output << " << data->id << "] deactivate" << endl; 
	#endif
	wextra.file.close();
}

void text_output::special_pause()
{
	#ifdef DEBUG_OUTPUT
	cout << "[text_output << " << data->id << "] pause" << endl; 
	#endif
}
