/*
	output: general output device and data definition 
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"

#include "audio_buffer.hxx"
#include "output.hxx"

#include "out_dummy.hxx"

#ifdef OUT_OSS
	#include "out_oss.hxx"
#endif

#ifdef OUT_MME
	#include "out_mme.hxx"
#endif

#ifdef OUT_ALSA
	#include "out_alsa.hxx"
#endif

#include "out_text.hxx"
#include "out_raw_s16.hxx"

output_device* new_output_device(const string type, output_data& data)
{
	if(type == dummy_output::t) return new dummy_output(&data, buffer_samples);
	else if(type == text_output::t) return new text_output(&data, buffer_samples);
	else if(type == raw_s16_output::t) return new raw_s16_output(&data, buffer_samples);
	#ifdef OUT_OSS
	else if(type.find(oss_output::t,0) == 0) return new oss_output(&data, buffer_samples, type); //choose serial/thread mode based on type
	#endif
	#ifdef OUT_MME
	else if(type == mme_output::t) return new mme_output(&data, buffer_samples);
	#endif
	#ifdef OUT_ALSA
	else if(type.find(alsa_output::t,0) == 0) return new alsa_output(&data, buffer_samples, type); //choose serial/thread mode based on type
	#endif
	else return NULL;
}
