#ifndef DMD_H_OUTPUT
#define DMD_H_OUTPUT

/*
	output: general output device and data definition (see also output_base)
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
	
	The generic (virtual) output class for DerMixD.
	
	Logic shall be here.
	Wetwork shall be in the special device.
	
	Multithreading of playback is taking care of; missing is the possibly wanted multithreading of loading...
	
	construct: open device/file/stream...
	play: put that buffer on tape!
	destruct: close device/file/stream...
	
	for deep sleep pausing of one can close devices and reopen them on unpause... 
*/

#include "thread_container.hxx"

typedef struct writerdata: public thread_data
{
	audio_buffer buffer;
	sem_t insem, outsem;
	volatile int error;
	void *extra; //some special data
	int id;
} writerdata;

class output_data
{
	public:
		audio_format volatile format;
		string filename;
		int id;
		unsigned long position; //since we won't seek (?), a float value may be more appropriate for it won't overflow on long sessions... if I really intend to use the output position sometime
		
		output_data(){ reset(); };
		void reset()
		{
			position = 0;
			filename = "";
		};
};

class output_device: public thread_container
{
	protected:
		
		output_data* data;
		writerdata wdat;
		pthread_t wthread; //the writer thread
		int good_writer_error; //the non-error value for writer
		bool playing;
		bool active;
		bool on_line;
		bool have_sems;
		
		void prebuild(); //initialize stuff so that destroy will work even without build
		void build(output_data* odi, unsigned int bufsize, void* (*thewriter)(void*));
		virtual void endgame();
		virtual void destroy();
		
		//these and type() need implementation in the final device
		virtual bool special_init() = 0;
		virtual void special_exit() = 0;
		void* (*writer)(void*); //define a local non-class function and give it as parent constuctor argument
		virtual bool special_activate() = 0;
		virtual void special_deactivate() = 0;
		virtual void special_pause() = 0;
		virtual void special_write();
		
	public:
		audio_buffer buffer;
		//this constructor needs
		output_device(){};
		virtual ~output_device(){};
		bool activate(); //open, wake up, ...
		void deactivate(); //close it, suspend, ...
		virtual void play();
		void load(comm_data* cd, const string& filename);
		void pause(); //make sure the buffer gets cleared because there won't be following any writes too soon (for OSS, but  apparently not working)
		bool online();
		bool is_active(); //this returning false after activate() means trouble 
		bool is_playing();
		virtual bool resize_buffer(unsigned int s);
		
		virtual string type() = 0;
		
		string problem;
};

output_device* new_output_device(const string type, output_data &data);

#endif
