/*
	param: some parameters in the environment...
	
	part of DerMixD
	(c)2005-7 Thomas Orgis, licensed under GPLv2
*/

//not depending on dermixd's common header, this stuff can easily tranfserred to other projects or into an independent library
#include <cstdlib>
#include <cstdio>
#include <string>
#include <vector>
using namespace std;

#include "param.hxx"
#include "debug.hxx"

param_space::param_space()
{
	construct("param_space");
	maxlen = 0;
	separator = "--";
	program = "";
	the_lock = NULL;
	define("help","no","give help");
}

param_space::~param_space()
{
	destbegin("param_space");
	for(vector<parm_group*>::iterator g = pgv.begin(); g != pgv.end(); ++g)
	delete (*g);
	pgv.clear();
	single();
	destend("param_space");
}

string& param_space::val(const string group, const string name)
{
	parm* p = NULL;
	if(get_parm_struct(group,name,p) == 0) return p->value;
	else return dummy;
}

bool param_space::concurrent()
{
	if(the_lock == NULL)
	{
		the_lock = new pthread_mutex_t;
		if(the_lock != NULL)
		{
			if(pthread_mutex_init(the_lock,NULL) == 0) return true;
			else { delete the_lock; the_lock = NULL; return false; }
		}
		else return false;
	}
	return true;
}

void param_space::single()
{
	if(the_lock != NULL)
	{
		pthread_mutex_destroy(the_lock);
		delete the_lock;
		the_lock = NULL;
	}
}

void param_space::lock()
{
	if(the_lock != NULL) pthread_mutex_lock(the_lock);
}

void param_space::unlock()
{
	if(the_lock != NULL) pthread_mutex_unlock(the_lock);
}

parm_group* param_space::get_group(const string name)
{
	for(vector<parm_group*>::iterator i = pgv.begin(); i != pgv.end(); ++i)
	{
		if((*i)->name == name) return *i;
	}
	return NULL;
}

int param_space::get_parm_struct(const string group, const string name, parm* &pmp)
{
	parm_group* grp = get_group(group);
	if( grp != NULL )
	{
		if( (pmp = grp->get(name)) != NULL ) return 0; //found
		else
		{
			error("no param \"%s\" in group \"%s\"", name.c_str(), group.c_str());
			return -2;
		}
	}
	else
	{
		error("no group \"%s\"", group.c_str());
		return -1;
	}
}

void param_space::define(const string group, const string name, const string value, const string help)
{
	lock();
	parm_group* grp = get_group(group);
	if(grp == NULL)
	{
		pgv.push_back(new parm_group(group));
		grp = pgv.back();
	}
	parm* p = grp->get(name);
	if(p == NULL)
	{
		grp->parms.push_back(new parm);
		p = grp->parms.back();
	}
	else error("Redefining \"%s\" in group \"%s\" - the coder is serious?", name.c_str(), group.c_str());
	unsigned int nl = group.length() + name.length() + 1;
	if(nl > maxlen) maxlen = nl;
	p->name = name;
	p->help = help;
	p->value = value;
	unlock();
}

bool param_space::get(const string group, const string name, string& value)
{
	parm* p = NULL;
	lock();
	int error = get_parm_struct(group,name,p);
	if(error == 0) value = p->value;
	unlock();
	return error == 0;
}

bool param_space::set(const string group, const string name, const string value)
{
	parm* p = NULL;
	lock();
	int error = get_parm_struct(group,name,p);
	if(error == 0) p->value = value;
	unlock();
	return error == 0;
}

void param_space::help(FILE *out)
{
	fprintf(out, "\nParameter input: group.name=value; you can omit group \"main\"\n");
	fprintf(out, "\"%s\" stops parameter parsing (for file names or similar following)\n", separator.c_str());
	fprintf(out, "Example: program name=value main.another=anothervalue -- file1 file2\n\n");
	fprintf(out, "The parameters:\n");
	lock();
	for(vector<parm_group*>::iterator g = pgv.begin(); g != pgv.end(); ++g)
	for(vector<parm*>::iterator p = (*g)->parms.begin(); p != (*g)->parms.end(); ++p)
	{
		fprintf(out, "%s.%s:", (*g)->name.c_str(), (*p)->name.c_str());
		for(size_t j = 0; j < (maxlen - (*g)->name.length() - (*p)->name.length()); ++j) fprintf(out, " ");

		fprintf(out, "%s\t[%s]\n", (*p)->help.c_str(), (*p)->value.c_str());
	}
	unlock();
}

bool param_space::parse(const string arg)
{
	unsigned int i = 0;
	unsigned int j = 0;
	bool ret = false;
	string group = "";
	string name = "";
	string value = "";
	for(i = 0; i < arg.length(); ++i)
	{
		if(j == 0)
		{
			if(arg[i] == '.')
			{
				if(i == 0) group = "main";
				else group = arg.substr(0,i);
				j = i+1; //just after the dot
			}
		}
		
		if(arg[i] == '=')
		{
			if(i != j)
			{
				name = arg.substr(j,i-j);
				value = arg.length() > (i+1) ? arg.substr(i+1) : "";
			}
			if(j == 0) group = "main";
			break;
		}
	}
	if(group.empty() || name.empty()) error("error parsing %s", arg.c_str());
	else
	{
		parm* p = NULL;
		lock();
		int error = get_parm_struct(group,name,p);
		if(error == 0) { p->value = value; ret = true; }
		unlock();
	}
	return ret;
}

bool param_space::parse(int &argc, char **&argv, const bool parse_program)
{
	if(parse_program && argc > 0)
	{
		program = argv[0];
		--argc;
		++argv;
	}
	int oldc = argc;
	for(int c = 0; c < oldc; ++c)
	{
		string w = argv[0];
		debug("parsing arg %i: %s", c, w.c_str());
		if(w == separator) // skip separator and stop
		{
			++argv;	--argc;
			break;
		}
		// if there's no "=" we assume it's no parameter
		if(w.find("=") == string::npos) break;
		// now try to parse
		if(!parse(w)) return false;
		// we're done with this one
		++argv; --argc;
	}
	return true;
}

bool param_space::as_bool(const string group, const string name)
{
	const char *truism[] = { "true", "on", "y", "yes", "ya", "j", "ja", "please", "yup", "Alrighty then!", NULL };
	string val;
	if(get(group,name,val))
	{
		// any non-null number is true
		if(as_double(group, name) != 0) return true;
		// any entry of the positive list is true
		int t = -1;
		while(truism[++t] != NULL)
		if( strcasecmp(val.c_str(), truism[t]) == 0 ) return true;
	}
	return false;
}

unsigned int param_space::as_uint(const string group, const string name)
{
	string val;
	int temp = 0;
	if(get(group,name,val)) temp = atoi(val.c_str());
	if(temp > 0) return (unsigned int) temp;
	else return 0;
}

int param_space::as_int(const string group, const string name)
{
	string val;
	if(get(group,name,val)) return atoi(val.c_str());
	else return 0;
}

long int param_space::as_lint(const string group, const string name)
{
	string val;
	if(get(group,name,val)) return atol(val.c_str());
	else return 0;
}

unsigned long int param_space::as_ulint(const string group, const string name)
{
	string val;
	long int temp = 0;
	if(get(group,name,val)) temp = atol(val.c_str());
	if(temp > 0) return (unsigned long int) temp;
	else return 0;
}

size_t param_space::as_size(const string group, const string name)
{
	// This is paranoid, but _has_ long to be <= size_t?
	unsigned long tmp = as_ulint(group, name);
	if(tmp <= ((size_t)-1)) return (size_t) tmp;
	else return 0;
}

double param_space::as_double(const string group, const string name)
{
	string val;
	if(get(group,name,val)) return atof(val.c_str());
	else return 0;
}

bool param_space::is_equal(const string group, const string name, const string value)
{
	string val;
	if(get(group,name,val)) return val == value;
	else return false;
}

const char *param_space::c_str(const string group, const string name)
{
	parm* p;
	const char* val = "";  // could return NULL ... am I that strict?
	lock();
	int error = get_parm_struct(group,name,p);
	if(error == 0) val = p->value.c_str();
	unlock();
	return val;
}

//implementation of sub types

parm::parm(const string &n, const string &h, const string &v)
{
	name = n;
	help = h;
	value = v;
}

parm::parm()
{
	construct("parm");
}

parm::~parm()
{
	destruct("parm");
}

parm_group::parm_group(const string &n)
{
	construct("parm_group");
	name = n;
}

parm_group::~parm_group()
{
	destbegin("parm_group");
	for(vector<parm*>::iterator i = parms.begin(); i != parms.end(); ++i)
	delete (*i); 	
	parms.clear();
	destend("parm_group");
}

parm* parm_group::get(const string &n)
{
	for(vector<parm*>::iterator i = parms.begin(); i != parms.end(); ++i)
	{
		if((*i)->name == n) return *i;
	}
	return NULL;
}
