/*
	peers: list of communicators that are available for peer2peer communication.

	part of DerMixD
	(c)2008 Thomas Orgis, licensed under GPLv2
*/
#include "common.hxx"
#include "peers.hxx"
#include "socket_writer_data.hxx"

#define ME "peers"

peers::peers()
{
	construct(ME);
	pthread_mutex_init(&lock, NULL);
	sem_init(&readers, 0 , 0); // failures???
}

peers::~peers()
{
	destbegin(ME);
	sem_destroy(&readers);
	pthread_mutex_destroy(&lock);
	destend(ME);
}

bool peers::message(action &act)
{
	bool ret = false;
	reading();
	for(vector<one_peer>::iterator s = them.begin(); s != them.end(); ++s)
	if(s->name == act.strings[1])
	{
		string* ms = new string;
		act.msg(act.strings[2], *ms);
		s->data->write(ms);
		ret=true;
		break;
	}
	done_reading();
	return ret;
}

void peers::list(action &act, socket_writer_data &wr)
{
	vector<string*>* ah = new vector<string*>;
	reading();

	ah->push_back(new string);
	act.begin(*ah->back());

	for(vector<one_peer>::iterator s = them.begin(); s != them.end(); ++s)
	{
		ah->push_back(new string("\t"));
		*ah->back() += s->name;
		*ah->back() += "\t";
		*ah->back() += s->description;
	}

	ah->push_back(new string);
	act.end(*ah->back());

	done_reading();
	wr.writes(ah);
}


void peers::writing()
{
	sdebug("going to lock and wait until all readers gone");
	pthread_mutex_lock(&lock);
	// So what... nasty busy waiting, trying to cover that with sched_yields().
	// We do wait attempts until the semaphore is zero,
	// but it should not be zeroed by this thread, so on success, it's increased again.
	// The socketeer threads are responsible for the "real" sem_wait().
	while(sem_trywait(&readers) == 0)
	{
		sem_post(&readers);
		sched_yield();
		sched_yield();
		sched_yield();
	}
	sdebug("can write now");
}

void peers::done_writing()
{
	sdebug("done writing");
	pthread_mutex_unlock(&lock);
}

bool peers::add(socket_writer_data* peerdata, const string &name, const string &descr)
{
	vector<one_peer>::iterator i;
	reading();
	/* Look if name already taken. */
	for(i=them.begin(); i!=them.end(); ++i)
	if(i->name == name) break;
	done_reading();

	if(i == them.end())
	{
		writing();
		them.push_back(one_peer());
		them.back().data = peerdata;
		them.back().name = name;
		them.back().description = descr;
		done_writing();
		return true;
	}
	else return false;
}

bool peers::remove(const string &name)
{
	bool ret = false;
	writing();
	for(vector<one_peer>::iterator i=them.begin(); i!=them.end(); ++i)
	if(i->name == name)
	{
		them.erase(i);
		ret = true;
		break;
	}
	done_writing();
	return ret;
}

void peers::eliminate(socket_writer_data *data)
{
	vector<one_peer>::iterator i;
	writing();
	/* Trying to delete _multiple_ occurences in a vector... */
	i = them.begin();
	while(i!=them.end())
	{
		if(i->data == data) i = them.erase(i);
		else ++i;
	}
	done_writing();
}

void peers::eliminate_all()
{
	writing();
	them.clear();
	done_writing();
}
