/*
	scanner: Scanner infrastructure and simple scanner implementations.

	part of DerMixD
	(c)2008 Thomas Orgis, licensed under GPLv2
*/

#include <math.h>
#include "common.hxx"
#include "scanner.hxx"
#include "scan_power.hxx"

scanner* create_scanner(const string &key, volatile audio_format *f)
{
	scanner* ret = NULL;
	if     (key == "length") ret = new length_scanner();
	else if(key == "peak")   ret = new peak_scanner();
	else if(key == "format") ret = new format_scanner();
	else if(key == "power")  ret = new power_scanner(f);
	else ret = new noname_scanner(key);

	ret->fmt = f;
	return ret;
}

scanner::~scanner(){} // Sense... something with vtable?

noname_scanner::noname_scanner(const string nm)
{
	name = nm;
}

void noname_scanner::report(tstring &rs)
{
	rs = "error: no scanner named " + name;
}


void peak_scanner::scan(audio_buffer &sb)
{
	for(size_t s=0; s < sb.fill*sb.channels; ++s)
	{
		audio_type cur = audio_abs(sb.ptr[s]);
		if(cur > peak) peak=cur;
	}
}

void peak_scanner::report(tstring &rs)
{
	double vol = audio_vol(peak);
	rs.sprintf("%g peak", vol);
	if(vol > 0) rs.sprintf(", %g dBFS", audio_vol2dbfs(vol));
}

void length_scanner::scan(audio_buffer &sb)
{
	length += sb.fill;
}

void length_scanner::report(tstring &rs)
{
	rs.sprintf("%lu samples", length);
}

void length_scanner::update_input(input_data *data)
{
	data->length = length;
}

void format_scanner::report(tstring &rs)
{
	rs.sprintf("%lu Hz, %u ch", fmt->rate, fmt->channels);
}
