/*
	socket_writer_data: data structure and write methods acting as interface for socker_writer thread (that is defined elsewhere)
	
	part of DerMixD
	(c)2005 Thomas Orgis, licensed under GPLv2
	
	depends on netcomm.h to provide actual socket reading/writing
*/

#include "common.hxx"
#include "socket_writer_data.hxx"
#include <sstream>

socket_writer_data::socket_writer_data()
{
	construct("socket_writer_data");
	is_end=false;
	sem_init(&semo, 0 , 0); // cannot fail with these parameters
	pthread_mutex_init(&lock, NULL);
}

socket_writer_data::~socket_writer_data()
{
	destbegin("socket_writer_data");
	prepare_flush(); // ensure we have all messages in old_messages.
	for(vector<swm>::iterator i = old_messages.begin(); i != old_messages.end(); ++i)
	{
		xdebug("destruct: deleting string at %p", (void*) i->msg);
		delete i->msg;
		sxdebug("destruct: done");
	}
	sem_destroy(&semo);
	destend("socket_writer_data");
}

void socket_writer_data::end()
{
	pthread_mutex_lock(&lock);
	is_end = true;
	sem_post(&semo);
	pthread_mutex_unlock(&lock);
}

void socket_writer_data::write(string *msg, bool spy)
{
	pthread_mutex_lock(&lock);
	if(msg == NULL || msg->c_str() == NULL)
	{
		serror("Tried to push null pointer! Must be a programming error ... triggering a segfault (look at the coredump...");
		printf("%s\n", msg->c_str());
	}
	else messages.push_back(swm(msg, spy));

	sem_post(&semo);
	pthread_mutex_unlock(&lock);
}

void socket_writer_data::writes(vector<string*> *msgs)
{
	pthread_mutex_lock(&lock);
	for(vector<string*>::iterator i = msgs->begin(); i != msgs->end(); ++i)
	{
		if(*i == NULL || (*i)->c_str() == NULL)
		{
			serror("Tried to push null pointer! Must be a programming error ... triggering a segfault (look at the coredump...");
			printf("%s\n", (*i)->c_str());
		}
		else
		{
			messages.push_back(swm(*i, false));
			sem_post(&semo);
		}
	}
	pthread_mutex_unlock(&lock);
	//the strings are deleted by the writer anyway... is the vector clear()ed on destruct? Think so.
	//or not? valgrind says that there are 4-byte blocks lost...
	//valgrind doesn't change it's mind when I do clear here... I really think the destructor does it already. And valgrind somehow cannot cope with it.
	//msgs->clear();
	delete msgs;
}

void socket_writer_data::flush(socket_buffer *sb, spies *bnd)
{
	for(vector<swm>::iterator i = old_messages.begin(); i != old_messages.end(); ++i)
	{
		sb->writeln(i->msg);
		// Now give all spies their info... all but myself!
		if(bnd && !i->spied) bnd->hand_over(sb->spyoutprefix + *i->msg, this);

		delete i->msg;
	}
	old_messages.clear();
}

void socket_writer_data::prepare_flush()
{
	pthread_mutex_lock(&lock);
	for(vector<swm>::iterator i = messages.begin(); i != messages.end(); ++i)
	old_messages.push_back(*i);

	messages.clear();
	sem_zero(&semo);
	pthread_mutex_unlock(&lock);
}

void socket_writer_data::write_copy(const string &msg)
{
	string* tmsg = new string;
	*tmsg = msg;
	write(tmsg);
}

void socket_writer_data::write_copies(vector<string*> &msgs)
{
	pthread_mutex_lock(&lock);
	for(vector<string*>::iterator i = msgs.begin(); i != msgs.end(); ++i)
	{
		if(*i == NULL || (*i)->c_str() == NULL)
		{
			serror("Tried to push null pointer! Must be a programming error ... triggering a segfault (look at the coredump...");
			printf("%s\n", (*i)->c_str());
		}
		else
		{
			// allocation failures???
			string *tmsg = new string;
			*tmsg = **i;
			messages.push_back(swm(tmsg, false));
			sem_post(&semo);
		}
	}
	pthread_mutex_unlock(&lock);
}
