/*
	spies: list of communicators spying on all communication

	part of DerMixD
	(c)2008 Thomas Orgis, licensed under GPLv2
*/
#include "common.hxx"
#include "socket_writer_data.hxx"

#define ME "spies"

spies::spies()
{
	construct(ME);
	active = false;
	pthread_mutex_init(&lock, NULL);
	sem_init(&readers, 0 , 0); // failures???
}

spies::~spies()
{
	destbegin(ME);
	sem_destroy(&readers);
	pthread_mutex_destroy(&lock);
	destend(ME);
}

void spies::hand_over(const string &msg, socket_writer_data *collaborateur)
{
	for(vector<socket_writer_data*>::iterator s = them.begin(); s != them.end(); ++s)
	if(*s != collaborateur){ string *cp = new string(msg); (*s)->write(cp, true); }
}

void spies::writing()
{
	sdebug("going to lock and wait until all readers gone");
	pthread_mutex_lock(&lock);
	// So what... nasty busy waiting, trying to cover that with sched_yields().
	// We do wait attempts until the semaphore is zero,
	// but it should not be zeroed by this thread, so on success, it's increased again.
	// The socketeer threads are responsible for the "real" sem_wait().
	while(sem_trywait(&readers) == 0)
	{
		sem_post(&readers);
		sched_yield();
		sched_yield();
		sched_yield();
	}
	sdebug("can write now");
}

void spies::done_writing()
{
	sdebug("done writing");
	pthread_mutex_unlock(&lock);
}

void spies::send(socket_writer_data* spy)
{
	vector<socket_writer_data*>::iterator i;
	reading();
	for(i=them.begin(); i!=them.end(); ++i)
	if(*i == spy) break;
	done_reading();

	if(i == them.end())
	{
		writing();
		them.push_back(spy);
		active = true;
		done_writing();
	}
}

void spies::eliminate(socket_writer_data *spy)
{
	writing();
	for(vector<socket_writer_data*>::iterator i=them.begin(); i!=them.end(); ++i)
	if(*i == spy)
	{
		them.erase(i);
		break;
	}
	if(them.empty()) active = false;
	done_writing();
}

void spies::eliminate_all()
{
	writing();
	them.clear();
	active = false;
	done_writing();
}
